--  pack.ex : lossless data compressor
--  jiri babor
--  jbabor@paradise.net.nz
--  00-01-29
--  version 1.10
----------------------------------------------------------------------
--  usage:  ex  pack inFilename  outFilename
--  where   inFilename : data file to be compressed
--          outFilename: output file with compressed data
----------------------------------------------------------------------
--  Algorithm:
--      As the bytes are read from the input file, the last pair is
--      compared with all adjacent pairs in the *last* 256 elements
--      of alredy processed data. If a match is found, the last pair
--      is shrunk to a single token containing the location of the
--      matching pair modulo 256, and the ninth bit of this token is
--      set to distiguish it from plain data. In Euphorian terms:
--
--          token = and_bits(index_of_matched_pair - 1, 255) + 256
--
--      At the end, ninth bits of *all* elements are collected,
--      converted into bytes and appended to the compressed output.
----------------------------------------------------------------------

include machine.e
include file.e

object junk
sequence b,cl
atom t
integer bp,next,fi,fo,ilen,sp

procedure help(sequence error_message)
    puts(1,"Error :  " & error_message & "\n")
    puts(1,"Syntax:  ex  pack  inFilename  outFilename\n")
    abort(1)
end procedure

procedure append_bits()
    -- convert flag bits to bytes and append them to end of output
    sequence bits
    integer m
    bits = b > #FF
    m = remainder(length(bits),8)
    if m then
        m = 8-m
        bits &= repeat(0, m)
    end if
    for i=1 to length(bits) by 8 do
        b &= bits_to_int(bits[i..i+7])
    end for
end procedure

-- main ------------------------------------------------------------------------

cl=command_line()
if length(cl)!=4 then
    help("Incorrect number of arguments!")
end if
fi=open(cl[3],"rb")
if fi=-1 then
    help("Input file " & cl[3] & " not found!")
end if

junk=seek(fi,-1)                -- go to end of input file
ilen=where(fi)                  -- get length of input file in bytes
junk=seek(fi,0)                 -- go back to beginning of input file
tick_rate(100)                  -- better timer resolution
t=time()                        -- start the clock
b={}                            -- processed (compressed) data sequence
bp=0                            -- init data pointer
sp=1                            -- init slider pointer
next=getc(fi)                   -- get very first byte
while next!=-1 do               -- while not end of input file
    while sp<bp-1 do            -- shrink routine
        if b[sp]=b[bp] then     -- artificial short-circuit
            if b[sp+1]=next then
                next=and_bits(sp-1,#FF)+#100
                bp-=1
                b=b[1..bp]
            end if
        end if
        sp += 1                 -- update search zone pointer
    end while
    b=append(b,next)            -- add to output
    bp+=1                       -- update data pointer
    next=getc(fi)               -- retrive the next byte from input file
    sp=bp-#101                  -- go back to beginning of search zone
    if sp<1 then sp=1 end if    -- but not past the beginning of file
end while
close(fi)                       -- close input file

append_bits()                   -- add flags to output
fo=open(cl[4],"wb")             -- open output file
puts(fo,b)                      -- write out
close(fo)                       -- close output file
t=time()-t                      -- stop clock

printf(1,"Compression time  %.2f  seconds\nShrunk from %d to %d bytes  " &
         "(%.2f%% of original)\n", {t,ilen,length(b),100*length(b)/ilen})
